import struct
import numpy as np
from hashlib import md5

from flipnote.schema import convertKWZFSIDToPPM

FRAMERATES = [0.2, 0.5, 1, 2, 4, 6, 8, 12, 20, 24, 30]

PALETTE = [(0xff, 0xff, 0xff),
           (0x10, 0x10, 0x10),
           (0xff, 0x10, 0x10),
           (0xff, 0xe7, 0x00),
           (0x00, 0x86, 0x31),
           (0x00, 0x38, 0xce),
           (0xff, 0xff, 0xff)]

# table1 - commonly occurring line offsets
TABLE_1 = np.array([0x0000, 0x0CD0, 0x19A0, 0x02D9, 0x088B, 0x0051, 0x00F3, 0x0009,
                    0x001B, 0x0001, 0x0003, 0x05B2, 0x1116, 0x00A2, 0x01E6, 0x0012,
                    0x0036, 0x0002, 0x0006, 0x0B64, 0x08DC, 0x0144, 0x00FC, 0x0024,
                    0x001C, 0x0004, 0x0334, 0x099C, 0x0668, 0x1338, 0x1004, 0x166C], dtype=np.uint16)

# table2 - commonly occurring line offsets, but the lines are shifted to the left by one pixel
TABLE_2 = np.array([0x0000, 0x0CD0, 0x19A0, 0x0003, 0x02D9, 0x088B, 0x0051, 0x00F3,
                    0x0009, 0x001B, 0x0001, 0x0006, 0x05B2, 0x1116, 0x00A2, 0x01E6,
                    0x0012, 0x0036, 0x0002, 0x02DC, 0x0B64, 0x08DC, 0x0144, 0x00FC,
                    0x0024, 0x001C, 0x099C, 0x0334, 0x1338, 0x0668, 0x166C, 0x1004], dtype=np.uint16)

ADPCM_STEP_TABLE = np.array([7, 8, 9, 10, 11, 12, 13, 14, 16, 17,
                             19, 21, 23, 25, 28, 31, 34, 37, 41, 45,
                             50, 55, 60, 66, 73, 80, 88, 97, 107, 118,
                             130, 143, 157, 173, 190, 209, 230, 253, 279, 307,
                             337, 371, 408, 449, 494, 544, 598, 658, 724, 796,
                             876, 963, 1060, 1166, 1282, 1411, 1552, 1707, 1878, 2066,
                             2272, 2499, 2749, 3024, 3327, 3660, 4026, 4428, 4871, 5358,
                             5894, 6484, 7132, 7845, 8630, 9493, 10442, 11487, 12635, 13899,
                             15289, 16818, 18500, 20350, 22385, 24623, 27086, 29794, 32767], dtype=np.int16)

# index table for 2-bit samples
ADPCM_INDEX_TABLE_2 = np.array([-1, 2, -1, 2], dtype=np.int8)

# index table for 4-bit samples
ADPCM_INDEX_TABLE_4 = np.array([-1, -1, -1, -1, 2, 4, 6, 8,
                                -1, -1, -1, -1, 2, 4, 6, 8], dtype=np.int8)


class Parser:
    def __init__(self, buffer=None, processing_frames=True):
        # layer output buffers
        self.buffer = None
        self.size = 0

        self.sections = {}

        self.is_folder_icon = False

        self.frame_offsets = []
        self.frame_meta = []
        self.frame_count = 0

        self.track_frame_speed = 0

        self.layer_pixels = np.zeros((3, 240, 40), dtype="V8")

        self.thumb_index = 0
        self.frame_speed = 0
        self.layer_visibility = [False, False, False]

        self.track_lengths = None

        self.track_meta = None

        self.framerate = None

        self.meta = None

        self.prev_decoded_frame = 0

        self.LINE_TABLE = np.zeros(6561, dtype="V8")
        self.TABLE_3 = np.zeros(6561, dtype=np.uint16)

        # initial values for read_bits()
        self.bit_index = 16
        self.bit_value = 0

        if buffer:
            self.load(buffer, processing_frames)

    @classmethod
    def open(cls, path):
        with open(path, "rb") as buffer:
            return cls(buffer)

    def load(self, buffer, processing_frames):
        self.buffer = buffer

        # lazy way to get file length:
        # seek to the end (ignore signature), get the position, then seek back to the start
        self.buffer.seek(0, 2)
        self.size = self.buffer.tell() - 256
        self.buffer.seek(0, 0)

        # build list of section offsets + lengths
        offset = 0
        while offset < self.size:
            self.buffer.seek(offset)
            magic, length = struct.unpack("<3sxI", self.buffer.read(8))
            self.sections[str(magic, 'ascii')] = {"offset": offset, "length": length}
            offset += length + 8

        # read file header -- not present in folder icons
        if "KFH" in self.sections:
            self.decode_meta()
        else:
            self.is_folder_icon = True
            self.frame_count = 1

        # read sound data header -- not present in comments or icons
        if "KSN" in self.sections:
            self.buffer.seek(self.sections["KSN"]["offset"] + 8)
            self.track_frame_speed = struct.unpack("<I", self.buffer.read(4))
            self.track_lengths = struct.unpack("<IIIII", self.buffer.read(20))
            self.meta.update(self.get_track_meta())

        if processing_frames:
            self.buffer.seek(self.sections["KMI"]["offset"] + 8)
            offset = self.sections["KMC"]["offset"] + 12

            # parse each frame meta entry
            # https://github.com/Flipnote-Collective/flipnote-studio-3d-docs/wiki/kwz-format#kmi-frame-meta
            for i in range(self.frame_count):
                meta = struct.unpack("<IHHH10xBBBBI", self.buffer.read(28))

                self.frame_meta.append(meta)
                self.frame_offsets.append(offset)

                offset += meta[1] + meta[2] + meta[3]

                self.prev_decoded_frame = -1

    def unload(self):
        self.buffer.close()
        self.size = 0
        self.sections = {}
        self.frame_count = 0
        self.thumb_index = 0
        self.frame_speed = 0
        self.layer_visibility = [False, False, False]
        self.is_folder_icon = False
        self.frame_meta = []
        self.frame_offsets = []
        self.track_frame_speed = 0
        self.track_lengths = [0, 0, 0, 0, 0]
        self.prev_decoded_frame = -1

    def read_bits(self, num):
        if self.bit_index + num > 16:
            next_bits = int.from_bytes(self.buffer.read(2), byteorder="little")
            self.bit_value |= next_bits << (16 - self.bit_index)
            self.bit_index -= 16

        mask = (1 << num) - 1
        result = self.bit_value & mask

        self.bit_value >>= num
        self.bit_index += num

        return result

    def gen_line_tables(self):
        # table3 - line offsets, but the lines are shifted to the left by one pixel
        index = 0
        for a in range(0, 2187, 729):
            for b in range(0, 729, 243):
                for c in range(0, 243, 81):
                    for d in range(0, 81, 27):
                        for e in range(0, 27, 9):
                            for f in range(0, 9, 3):
                                for g in range(0, 3, 1):
                                    for h in range(0, 6561, 2187):
                                        self.TABLE_3[index] = a + b + c + d + e + f + g + h
                                        index += 1

        # linetable - contains every possible sequence of pixels for each tile line
        index = 0
        for a in range(3):
            for b in range(3):
                for c in range(3):
                    for d in range(3):
                        for e in range(3):
                            for f in range(3):
                                for g in range(3):
                                    for h in range(3):
                                        self.LINE_TABLE[index] = bytes([b, a, d, c, f, e, h, g])
                                        index += 1

    @staticmethod
    def decode_filename(raw_filename):
        try:
            return raw_filename.decode("ascii").rstrip("\x00")
        except UnicodeDecodeError:
            # in some DSi Gallery notes, Nintendo messed up and included the
            # PPM-format filename without converting it
            mac, ident, edits = struct.unpack("<3s13sH", raw_filename[0:18])
            return "{0}_{1}_{2:03d}".format("".join(["%02X" % c for c in mac]), ident.decode("ascii"), edits)

    def get_thumbnail(self):
        self.buffer.seek(self.sections["KTN"]["offset"] + 12)
        return self.buffer.read(self.sections["KTN"]["length"])

    def has_audio_track(self, track_index):
        return self.track_lengths[track_index] > 0

    def get_audio_track_offset(self, track_index):
        # offset starts after sound header
        offset = self.sections["KSN"]["offset"] + 36
        for i in range(track_index):
            offset += self.track_lengths[i]
        return offset

    def get_audio_track(self, track_index):
        size = self.track_lengths[track_index]

        self.buffer.seek(self.get_audio_track_offset(track_index))

        return self.buffer.read(size)

    def get_track_digest(self, track_index):
        if self.has_audio_track(track_index):
            return md5(self.get_audio_track(track_index)).hexdigest()
        else:
            return None

    def decode_audio_track(self, track_index, step_index=0):
        size = self.track_lengths[track_index]

        self.buffer.seek(self.get_audio_track_offset(track_index))

        # create an output buffer with enough space for 60 seconds of audio at 16364 Hz
        output = np.zeros(16364 * 60, dtype="<u2")
        output_offset = 0

        predictor = 0
        diff = 0

        for byte in self.buffer.read(size):
            bit_pos = 0
            while bit_pos < 8:
                if step_index < 18 or bit_pos > 4:
                    sample = byte & 0x3

                    step = ADPCM_STEP_TABLE[step_index]
                    diff = step >> 3

                    if sample & 1:
                        diff += step
                    if sample & 2:
                        diff = -diff

                    predictor += diff
                    step_index += ADPCM_INDEX_TABLE_2[sample]

                    byte >>= 2
                    bit_pos += 2
                else:
                    sample = byte & 0xF

                    step = ADPCM_STEP_TABLE[step_index]
                    diff = step >> 3

                    if sample & 1:
                        diff += step >> 2
                    if sample & 2:
                        diff += step >> 1
                    if sample & 4:
                        diff += step
                    if sample & 8:
                        diff = -diff

                    predictor += diff
                    step_index += ADPCM_INDEX_TABLE_4[sample]

                    byte >>= 4
                    bit_pos += 4

                # clamp step index and diff
                step_index = max(0, min(step_index, 79))
                diff = max(-2048, min(diff, 2047))

                # add result to output buffer
                output[output_offset] = predictor * 16
                output_offset += 1

        # trim output array to the size of the track before returning
        return output[:output_offset]

    def decode_meta(self):
        self.buffer.seek(self.sections["KFH"]["offset"] + 12)
        creation_timestamp, modified_timestamp, app_version = struct.unpack("<III", self.buffer.read(12))
        root_author_id, parent_author_id, current_author_id = struct.unpack("<10s10s10s", self.buffer.read(30))
        root_author_name, parent_author_name, current_author_name = struct.unpack("<22s22s22s", self.buffer.read(66))
        root_filename, parent_filename, current_filename = struct.unpack("<28s28s28s", self.buffer.read(84))
        self.frame_count, self.thumb_index, flags, self.frame_speed, layer_flags = struct.unpack("<HHHBB",
                                                                                                 self.buffer.read(8))
        self.framerate = FRAMERATES[self.frame_speed]
        self.layer_visibility = [layer_flags & 0x1 == 0,  # Layer A
                                 (layer_flags >> 1) & 0x1 == 0,  # Layer B
                                 (layer_flags >> 2) & 0x1 == 0]  # Layer C

        self.meta = {
            "lock": flags & 0x1,
            "loop": (flags >> 1) & 0x01,
            "flags": flags,
            "layer_flags": layer_flags,
            "app_version": app_version,
            "frame_count": self.frame_count,
            "frame_speed": self.frame_speed,
            "thumb_index": self.thumb_index,
            "modified_timestamp": modified_timestamp + 946684800,
            "creation_timestamp": creation_timestamp + 946684800,
            "root_username": root_author_name.decode("utf-16").rstrip("\x00"),
            "root_fsid": root_author_id.hex(),
            "root_fsid_ppm": convertKWZFSIDToPPM(root_author_id.hex()),
            "root_filename": self.decode_filename(root_filename),
            "parent_username": parent_author_name.decode("utf-16").rstrip("\x00"),
            "parent_fsid": parent_author_id.hex(),
            "parent_fsid_ppm": convertKWZFSIDToPPM(root_author_id.hex()),
            "parent_filename": self.decode_filename(parent_filename),
            "current_username": current_author_name.decode("utf-16").rstrip("\x00"),
            "current_fsid": current_author_id.hex(),
            "current_fsid_ppm": convertKWZFSIDToPPM(current_author_id.hex()),
            "current_filename": self.decode_filename(current_filename)
        }

        return self.meta

    def get_track_meta(self):
        return {
            "bgm_used": self.has_audio_track(0),
            "se1_used": self.has_audio_track(1),
            "se2_used": self.has_audio_track(2),
            "se3_used": self.has_audio_track(3),
            "se4_used": self.has_audio_track(4),
            "bgm_digest": self.get_track_digest(0),
            "se1_digest": self.get_track_digest(1),
            "se2_digest": self.get_track_digest(2),
            "se3_digest": self.get_track_digest(3),
            "se4_digest": self.get_track_digest(4)
        }



from re import match
from textwrap import wrap
from base64 import b32decode

"""
This file contains functions relating to various schema within flipnotes.
"""

KWZ_FSID_trans = str.maketrans("CWMFJORDVEGBALKSNTHPYXQUIZ012345", "ABCDEFGHIJKLMNOPQRSTUVWXYZ234567")


def verifyPPMFSID(input_fsid):
    """
    Verifies that a PPM format FSID is valid via regex
    """
    return match("[0159][0-9A-F]{15}", input_fsid) is not None


def verifyKWZFSID(input_fsid):
    """
    Verifies that a KWZ format FSID is valid via regex
    """
    return match("(00|10|12|14)[0-9A-F]{14}[0159][0-9A-F](00)?", input_fsid) is not None


def verifyKWZFilename(file_name):
    """
    Verifies that a KWZ format file name is valid via regex
    """
    return match(r"[a-z0-5]{28}(\.kwz)?", file_name) is not None


def verifyPPMFilename(name):
    """
    Verifies that a PPM format file name is valid via regex
    """
    return match(r"[0-9A-F]{6}_[0-9A-F]{13}_[0-9]{3}(\.ppm)?", name) is not None


def verifyPPMFilesystemFilename(name):
    """
    Verifies that a PPM format filesystem file name is valid via regex
    Filenames in the filesystem (outside of file meta) have the first character used as a checksum
    """
    return match(r"[0-9A-Z][0-9A-F]{5}_[0-9A-F]{13}_[0-9]{3}(\.ppm)?", name) is not None


def convertKWZFSIDToPPM(input_fsid):
    """
    Converts KWZ format FSIDs to PPM format.
    KWZ format FSIDs must be 18 or 20 characters.
    Any invalid input will be returned without modification.
    - e.g. if a PPM format FSID is used as the input or the length is invalid
    """
    output_fsid = ""

    # Clean up input
    input_fsid = str(input_fsid).strip().upper()

    if verifyKWZFSID(input_fsid):
        # Trim the first byte of the FSID
        # FSIDs from KWZ files have an extra null(?) byte at the end, trim it if it exists
        if len(input_fsid) == 20:
            input_fsid = input_fsid[2:-2]
        else:
            input_fsid = input_fsid[2:]

        # Invert the FSID then split into byte sized chunks
        string_list = wrap(input_fsid[::-1], 2)

        # Invert each byte and append to the output string
        for i in range(len(string_list)):
            output_fsid += string_list[i][::-1]
    else:
        output_fsid = input_fsid

    return output_fsid


def convertPPMtoKWZ(input_fsid):
    """
    Converts PPM format FSIDs to KWZ format
    The first byte appears to be useless, as all versions (00, 10, 12, 14) refer to the same user
    This returns the FSID with 00 as the leading byte by default
    - Once/if the purpose of this byte is discovered, it will be modified
    The trailing null byte is also included
    Any invalid input will be returned without modification.
    - e.g. if a PPM format FSID is used as the input or the length is invalid
    """

    # Clean up input
    output_fsid = str(input_fsid).strip().upper()

    if verifyPPMFSID(input_fsid):
        # Invert the FSID then split into byte sized chunks
        string_list = wrap(input_fsid[::-1], 2)

        # Invert each byte and append to the output string
        for i in range(len(string_list)):
            output_fsid += string_list[i][::-1]

        output_fsid = "00" + output_fsid + "00"

    return output_fsid


def unpackKWZFilename(input_filename):
    """
    Decodes a KWZ format filename to a string of its decoded hex bytes.
    Any invalid input that doesn't match the filename regex will be returned without modification.
    """
    # Clean up input
    output_filename = str(input_filename).strip().upper()

    if verifyKWZFilename(output_filename):
        # Convert custom base-32 encoded string to the standard base-32 alphabet
        str(output_filename).translate(KWZ_FSID_trans)

        # Add padding to allow for decoding
        input_filename += "===="

        output_filename = b32decode(input_filename).hex().upper()
    else:
        # Return without modification if verification fails
        output_filename = input_filename

    return output_filename




from setuptools import setup, find_packages

with open("README.md", "r", encoding="utf-8") as readme:
    long_description = readme.read()

setup(name="flipnote",
      version="0.1.3",
      description="A Python library for Flipnote Studio (3D) files",
      long_description=long_description,
      long_description_content_type="text/markdown",
      author="Meemo",
      author_email="meemo4556@gmail.com",
      license="MIT",
      install_requires=[
          "numpy"
      ],
      url="https://github.com/meemo/flipnote.py",
      project_urls={
          "Bug Tracker": "https://github.com/meemo/flipnote.py/issues",
      },
      classifiers=[
          "Programming Language :: Python :: 3",
          "License :: OSI Approved :: MIT License",
          "Operating System :: OS Independent",
      ],
      package_dir={"": "src"},
      packages=find_packages(where="src"),
      python_requires=">=3.7")



import struct
import numpy as np
from datetime import datetime

# Flipnote speed -> frames per second
# Framrates are indexed from 1, so the first value is a placeholder
FRAMERATES = [0, 0.5, 1, 2, 4, 6, 12, 20, 30]

# Thumbnail bitmap RGB colors
THUMBNAIL_PALETTE = [(0xFF, 0xFF, 0xFF),
                     (0x52, 0x52, 0x52),
                     (0xFF, 0xFF, 0xFF),
                     (0x9C, 0x9C, 0x9C),
                     (0xFF, 0x48, 0x44),
                     (0xC8, 0x51, 0x4F),
                     (0xFF, 0xAD, 0xAC),
                     (0x00, 0xFF, 0x00),
                     (0x48, 0x40, 0xFF),
                     (0x51, 0x4F, 0xB8),
                     (0xAD, 0xAB, 0xFF),
                     (0x00, 0xFF, 0x00),
                     (0xB6, 0x57, 0xB7),
                     (0x00, 0xFF, 0x00),
                     (0x00, 0xFF, 0x00),
                     (0x00, 0xFF, 0x00)]

# Frame RGB colors
BLACK = (0x0E, 0x0E, 0x0E)
WHITE = (0xFF, 0xFF, 0xFF)
BLUE = (0x0A, 0x39, 0xFF)
RED = (0xFF, 0x2A, 0x2A)


class Parser:
    @classmethod
    def open(cls, path):
        f = open(path, "rb")
        return cls(f)

    def __init__(self, stream=None):
        self.lock = None
        self.thumb_index = None

        if stream:
            self.load(stream)

    def load(self, stream):
        self.stream = stream
        self.read_header()
        self.read_meta()
        self.read_animation_header()
        self.read_sound_header()
        self.layers = np.zeros((2, 192, 256), dtype=np.uint8)
        self.prev_layers = np.zeros((2, 192, 256), dtype=np.uint8)
        self.prev_frame_index = -1

    def unload(self):
        self.stream.close()

    def read_header(self):
        # decode header
        # https://github.com/pbsds/hatena-server/wiki/PPM-format#file-header
        self.stream.seek(0)
        magic, animation_data_size, sound_data_size, frame_count, version = struct.unpack("<4sIIHH",
                                                                                          self.stream.read(16))
        self.animation_data_size = animation_data_size
        self.sound_data_size = sound_data_size
        self.frame_count = frame_count + 1

    def read_filename(self):
        # Parent and current filenames are stored as:
        #  - 3 bytes representing the last 6 digits of the console's MAC address
        #  - 13-character string
        #  - uint16 edit counter
        mac, ident, edits = struct.unpack("<3s13sH", self.stream.read(18))
        # Filenames are formatted as <3-byte MAC as hex>_<13-character string>_<edit counter as a 3-digit number>
        # eg F78DA8_14768882B56B8_030
        return "{0}_{1}_{2:03d}".format("".join(["%02X" % c for c in mac]), ident.decode("ascii"), edits)

    def read_meta(self):
        # decode metadata
        # https://github.com/pbsds/hatena-server/wiki/PPM-format#file-header
        self.stream.seek(0x10)
        self.lock, self.thumb_index = struct.unpack("<HH", self.stream.read(4))
        self.root_author_name = self.stream.read(22).decode("utf-16").rstrip("\x00")
        self.parent_author_name = self.stream.read(22).decode("utf-16").rstrip("\x00")
        self.current_author_name = self.stream.read(22).decode("utf-16").rstrip("\x00")
        self.parent_author_id = "%016X" % struct.unpack("<Q", self.stream.read(8))
        self.current_author_id = "%016X" % struct.unpack("<Q", self.stream.read(8))
        self.parent_filename = self.read_filename()
        self.current_filename = self.read_filename()
        self.root_author_id = "%016X" % struct.unpack("<Q", self.stream.read(8))
        self.partial_filename = self.stream.read(8)  # not really useful for anything :/
        # timestamp is stored as the number of seconds since jan 1st 2000
        # we add 946684800 to convert this to a unix timestamp
        self.timestamp = datetime.fromtimestamp(struct.unpack("<I", self.stream.read(4))[0] + 946684800)

    def read_thumbnail(self):
        self.stream.seek(0xA0)
        bitmap = np.zeros((48, 64), dtype=np.uint8)
        for tile_index in range(0, 48):
            tile_x = tile_index % 8 * 8
            tile_y = tile_index // 8 * 8
            for line in range(0, 8):
                for pixel in range(0, 8, 2):
                    byte = ord(self.stream.read(1))
                    x = tile_x + pixel
                    y = tile_y + line
                    bitmap[y][x] = byte & 0x0F
                    bitmap[y][x + 1] = (byte >> 4) & 0x0F
        return bitmap

    def read_animation_header(self):
        self.stream.seek(0x06A0)
        table_size, unknown, flags = struct.unpack("<HHI", self.stream.read(8))
        # unpack animation flags
        self.layer_1_visible = (flags >> 11) & 0x01
        self.layer_2_visible = (flags >> 10) & 0x01
        self.loop = (flags >> 1) & 0x01
        # read offset table into a numpy array
        offset_table = np.frombuffer(self.stream.read(table_size), dtype=np.uint32)
        self.offset_table = [offset + 0x06A0 + 8 + table_size for offset in offset_table]

    def read_sound_header(self):
        # https://github.com/pbsds/hatena-server/wiki/PPM-format#sound-data-section
        # offset = frame data offset + frame data length + sound effect flags
        offset = 0x06A0 + self.animation_data_size + self.frame_count
        # account for multiple-of-4 padding
        if offset % 2 != 0:
            offset += 4 - (offset % 4)
        self.stream.seek(offset)
        bgm_size, se1_size, se2_size, se3_size, frame_speed, bgm_speed = struct.unpack("<IIIIBB", self.stream.read(18))
        self.frame_speed = 8 - frame_speed
        self.bgm_speed = 8 - bgm_speed
        self.framerate = FRAMERATES[self.frame_speed]
        self.bgm_framerate = FRAMERATES[self.bgm_speed]

    def is_frame_new(self, index):
        self.stream.seek(self.offset_table[index])
        return ord(self.stream.read(1)) >> 7 & 0x1

    @staticmethod
    def read_line_types(line_types):
        for index in range(192):
            line_type = line_types[index // 4] >> ((index % 4) * 2) & 0x03
            yield index, line_type

    def read_frame(self, index):
        # decode previous frames if needed
        if index != 0 and self.prev_frame_index != index - 1 and not self.is_frame_new(index):
            self.read_frame(index - 1)
        # copy the current layer buffers to the previous ones
        np.copyto(self.prev_layers, self.layers)
        self.prev_frame_index = index
        # clear current layer buffers by resetting them to 0
        self.layers.fill(0)
        # seek to the frame offset so we can start reading
        self.stream.seek(self.offset_table[index])
        # unpack frame header flags
        header = ord(self.stream.read(1))
        is_new_frame = (header >> 7) & 0x1
        is_translated = (header >> 5) & 0x3
        translation_x = 0
        translation_y = 0
        # if the frame is translated, we need to unpack the x and y values
        if is_translated:
            translation_x, translation_y = struct.unpack("<bb", self.stream.read(2))
        # read line encoding bytes
        line_types = [
            self.stream.read(48),
            self.stream.read(48),
        ]
        # loop through layers
        for layer in range(2):
            bitmap = self.layers[layer]
            for line, line_type in self.read_line_types(line_types[layer]):
                pixel = 0
                # no data stored for this line
                if line_type == 0:
                    pass
                # compressed line
                elif line_type == 1 or line_type == 2:
                    # if line type == 2, the line starts off with all the pixels set to 1
                    if line_type == 2:
                        for i in range(256):
                            bitmap[line][i] = 1
                    # unpack chunk usage
                    chunk_usage = struct.unpack(">I", self.stream.read(4))[0]
                    # unpack pixel chunks
                    while pixel < 256:
                        if chunk_usage & 0x80000000:
                            chunk = ord(self.stream.read(1))
                            for bit in range(8):
                                bitmap[line][pixel] = chunk >> bit & 0x1
                                pixel += 1
                        else:
                            pixel += 8
                        chunk_usage <<= 1
                # raw line
                elif line_type == 3:
                    # unpack pixel chunks
                    while pixel < 256:
                        chunk = ord(self.stream.read(1))
                        for bit in range(8):
                            bitmap[line][pixel] = chunk >> bit & 0x1
                            pixel += 1

        # frame diffing - if the current frame is based on the previous one, merge them by XORing their pixels
        # this is a big performance bottleneck...
        if not is_new_frame:
            # loop through lines
            for y in range(192):
                # skip to next line if this one falls off the top edge of the screen
                if y - translation_y < 0:
                    continue
                # stop once the bottom screen edge has been reached
                if y - translation_y >= 192:
                    break
                for x in range(256):
                    # skip to the next pixel if this one falls off the left edge of the screen
                    if x - translation_x < 0:
                        continue
                    #  stop diffing this line once the right screen edge has been reached
                    if x - translation_x >= 256:
                        break
                    # diff pixels with a binary XOR
                    self.layers[0][y][x] ^= self.prev_layers[0][y - translation_y][x - translation_x]
                    self.layers[1][y][x] ^= self.prev_layers[1][y - translation_y][x - translation_x]

        return self.layers

    def get_frame_palette(self, index):
        self.stream.seek(self.offset_table[index])
        header = ord(self.stream.read(1))
        paper_color = header & 0x1
        pen = [None,
               BLACK if paper_color == 1 else WHITE,
               RED,
               BLUE]
        return [WHITE if paper_color == 1 else BLACK,
                pen[(header >> 1) & 0x3],  # layer 1 color
                pen[(header >> 3) & 0x3]]  # layer 2 color

    def get_frame_pixels(self, index):
        layers = self.read_frame(index)
        pixels = np.zeros((192, 256), dtype=np.uint8)
        for y in range(192):
            for x in range(256):
                if layers[0][y][x] > 0:
                    pixels[y][x] = 1
                elif layers[1][y][x] > 0:
                    pixels[y][x] = 2
        return pixels
